#
# $Copyright (c) 2018 Veritas Technologies LLC.  All rights reserved. Veritas
# and the Veritas Logo are trademarks or registered trademarks of Veritas
# Technologies LLC or its affiliates in the U.S. and other countries. Other 
# names may be trademarks of their respective owners.
# 
# The Licensed Software and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software - Restricted Rights" and
# DFARS 227.7202, et seq. "Commercial Computer Software and Commercial Computer
# Software Documentation," as applicable, and any successor regulations, whether
# delivered by Veritas as on premises or hosted services. Any use, modification,
# reproduction release, performance, display or disclosure of the Licensed Software
# and Documentation by the U.S. Government shall be solely in accordance with the
# terms of this Agreement. $

# $Id: RemoteCopyAgent.pm,v 1.1.2.21 2014/08/26 15:29:33 dsathe Exp $

#highest-message-id 80 

package RemoteCopyAgent;
use lib '.'; #TODO: use symlink creation logic once RPO.pm file location is fixed
use ag_i18n_inc;
use ArrayRep;

require RPO;
@ISA = ("ArrayRep");

use Date::Parse;

my $SHOWRCOPY = "showrcopy";
my $SHOWRCOPY_TARGET = "showrcopy targets";
my $FAILOVER = "setrcopygroup failover";
my $MIGRATE = "setrcopygroup reverse";
my $STARTRCOPYGROUP = "startrcopygroup";
my $RESTORE = "setrcopygroup restore";
my $RECOVER = "setrcopygroup recover";
my $REVERT = "setrcopygroup revert";
my $SWAPROLES = "setrcopygroup reverse -natural";
my $FAILOVER_SEC_REV = "setrcopygroup reverse -local";
my $SYNCRCOPY = "syncrcopy";
my $SHOWDATE = "showdate";
my $SHOWHOST = "showhost -d";

%default = {
	storage_server =>'',
        user =>'',
	passwordfile=>'',
        group =>'',
	cli =>'',
        ssh_bin =>'',
        ssh_path_to_id_file =>'',
	auto_takeover => '',
	swap_roles => '',
	resname => '',
	target => '',
        group_status => '',
        role => '',
        mode => '',
	sync_status => '',
	target_status  => '',
	force_sync => '',
	usecli =>'',
};
 
 sub new {
 	my $proto = shift;
 	my $resname = shift;
 	my @attrs = @_;
 	my $class = ref($proto) || $proto;
 	my $self = { %default };
 	$self->{'resname'} = $resname;
 	bless($self, $class);
 	my $ret = $VCSAG_SUCCESS;
	$self->{'start_time'} = $^T;
 	
 	($ret, $self->{'storage_server'}) = VCSAG_GET_ATTR_VALUE ("StorageServer", -1, 1, @attrs);
 	exit 0 if ($ret != $VCSAG_SUCCESS);
 
 	($ret, $self->{'user'}) = VCSAG_GET_ATTR_VALUE ("UserName", -1, 1, @attrs);
 	#exit 0 if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'passwordfile'}) = VCSAG_GET_ATTR_VALUE ("PasswordFile", -1, 1, @attrs);
 	#exit 0 if ($ret != $VCSAG_SUCCESS);
 
	($ret, $self->{'cli'}) = VCSAG_GET_ATTR_VALUE ("CLIBinary", -1, 1, @attrs);
 	#exit 0 if ($ret != $VCSAG_SUCCESS);

 	($ret, $self->{'ssh_bin'}) = VCSAG_GET_ATTR_VALUE ("SSHBinary", -1, 1, @attrs);
 	#exit 0 if ($ret != $VCSAG_SUCCESS);
 
 	($ret, $self->{'ssh_path_to_id_file'}) = VCSAG_GET_ATTR_VALUE ("SSHPathToIDFile", -1, 1, @attrs);
 	#exit 0 if ($ret != $VCSAG_SUCCESS);
 
 	($ret, $self->{'group'}) = VCSAG_GET_ATTR_VALUE ("GroupName", -1, 1, @attrs);
 	exit 0 if ($ret != $VCSAG_SUCCESS);
 
 	($ret, $self->{'auto_takeover'}) = VCSAG_GET_ATTR_VALUE ("AutoTakeover", -1, 1, @attrs);
 	exit 0 if ($ret != $VCSAG_SUCCESS);
 	
 	($ret, $self->{'swap_roles'}) = VCSAG_GET_ATTR_VALUE ("SwapRoles", -1, 1, @attrs);
 	exit 0 if ($ret != $VCSAG_SUCCESS);
		
 	($ret, $self->{'force_sync'}) = VCSAG_GET_ATTR_VALUE ("ForceSync", -1, 1, @attrs);
 	exit 0 if ($ret != $VCSAG_SUCCESS);

 	if ( $^O !~ /^MSWin/i ){
		($ret, $self->{'computedrsla'}) = VCSAG_GET_ATTR_VALUE ("ComputeDRSLA", -1, 1, @attrs);
 		exit 0 if ($ret != $VCSAG_SUCCESS);
	}
	
	return $self;
 }
 
 #validates agent arguments
 #RC = 0, validation succeeded
 #RC = 1, validation failed
 sub validate_args {
 	my $self = shift;
 	my $ret = 0;
 	
 	if ($self->{'storage_server'} eq "") {
 		VCSAG_LOG_MSG("E", "Storage Server (name/IP) must be specified.", 1);
 		$ret = 1;
 	}
 	if ($self->{'ssh_bin'} eq "") {
 		if ($self->{'cli'} eq "") {
			VCSAG_LOG_MSG("E", "Either SSHBinary or CLIBinary must be specified.", 76);
 			$ret = 1;
		} else {
			$self->{'usecli'} = 1;
		}
 	}
 	if ($self->{'usecli'} == 0 && $self->{'user'} eq "") {
 		VCSAG_LOG_MSG("E", "If SSHBinary is configured, Storage server User must be specified.", 2);
 		$ret = 1;
 	}
 	if ($self->{'usecli'} == 0 && $self->{'ssh_path_to_id_file'} eq "") {
 		VCSAG_LOG_MSG("E", "If SSHBinary is configured, SSHPathToIDFile must be specified.", 4);
 		$ret = 1;
 	}
	if ($self->{'usecli'} == 0 && $self->{'passwordfile'} ne "") {
                VCSAG_LOG_MSG("E", "If SSHBinary is configured, PasswordFile should not be specified.", 78);
                $ret = 1;
        }
	if ($self->{'usecli'} == 1 && $self->{'passwordfile'} eq "") {
 		VCSAG_LOG_MSG("E", "If CLIBinary is configured, PasswordFile must be specified.", 77);
 		$ret = 1;
 	} 
	if ($self->{'usecli'} == 1 && ( $self->{'user'} ne  "" || $self->{'ssh_path_to_id_file'} ne "")) {
 		VCSAG_LOG_MSG("E", "If CLIBinary is configured, User or SSHPathToIDFile should not be specified.", 79);
 		$ret = 1;
 	} 
 	if ($self->{'group'} eq "") {
 		VCSAG_LOG_MSG("E", "GroupName must be specified.", 5);
 		$ret = 1;
 	}
 	if (($self->{'auto_takeover'} != 0) && ($self->{'auto_takeover'} != 1)) {
 		VCSAG_LOG_MSG("E", "AutoTakeover must have value of 0 or 1.", 6);
 		$ret = 1;
 	}
	if (($self->{'swap_roles'} != 0) && ($self->{'swap_roles'} != 1)) {
                VCSAG_LOG_MSG("E", "SwapRoles must have value of 0 or 1.", 7);
                $ret = 1;
        }
	if (($self->{'force_sync'} != 0) && ($self->{'force_sync'} != 1)) {
                VCSAG_LOG_MSG("E", "ForceSync must have value of 0 or 1.", 69);
                $ret = 1;
        }
 		
 	return $ret;
 }

# Init other values
sub init_values {
        my ($self) = @_;
	my $init = 0;

	my ($ret, $out) = $self->remote_copy_cmd($SHOWRCOPY);

	if ($ret != 0) {
		return $init;
	}

	my @out = split('\n', $out);
	while (@out) {

            $line = shift @out;
            if ($line =~ /Group Information/) {
                shift @out;
                my $status_line = shift @out;
                my $role_line = shift @out;
                my (@role_line) =  split(" ", $role_line);
                $group = $role_line[0]; 
		$self->{'target'} = $role_line[1];
                $self->{'group_status'} = $role_line[2];
                $self->{'role'} = $role_line[3];
                $self->{'mode'} = $role_line[4];
		$init = 1;
                last;
            }
	}
	#if ($self->{'mode'} ne "Sync") {
	#	VCSAG_LOG_MSG("E", "Agent supports Sync mode. Group is configured in mode $self->{'mode'}.", 8, "$self->{'mode'}");
	#	return 0;
	#}
	
	$ret = $self->get_sync_status();
	if ($ret != 0) {
                return 0;
        }

	$ret = $self->get_target_status();
	if ($ret != 0) {
                return 0;
        }

	if ($init == 1 ) {
		 VCSAG_LOGDBG_MSG(1, "Group Name :: $self->{'group'}, Target :: $self->{'target'}, Group Status :: $self->{'group_status'}, Role ::$self->{'role'},"
			." Mode :: $self->{'mode'}", 9, "$self->{'group'}","$self->{'target'}", "$self->{'group_status'}", "$self->{'role'}", "$self->{'mode'}");
	} 
 	return $init;
}

#get_sync_status : Gets sync status
sub get_sync_status {
	my $self = shift;
	my $found = 0;
	my ($ret, $out) = $self->remote_copy_cmd($SHOWRCOPY);
	if ($ret != 0) {
                return $ret; 
        }
	my @out = split('\n', $out);
	while (@out) {
        	$line = shift @out;
        	if ($line =~ /LocalVV/) {
                	$found = 1;
                	$line = shift @out;
			if ($line eq "") { 
				VCSAG_LOG_MSG("E", "Virtual volumes are not configured for RemoteCopy group $self->{'group'}.Please Configure volumes for RemoteCopy group.", 10, "$self->{'group'}");
                        	$ret = 1;
				last;
                	}
        	}
        	if (($found == 1) && ($line ne "")) {
                	my (@sync_line)  = split (" ", $line);
			$self->{'sync_status'} = $sync_line[4];
			if ($self->{'sync_status'} ne "Synced") {
				#Sync status for VV is not Synced return this status of whole group.
				VCSAG_LOGDBG_MSG(1, "LocalVV :: $sync_line[0] Sync Status is $self->{'sync_status'}.", 11, "$sync_line[0]", "$self->{'sync_status'}");
				last;
        		}
		} 
	}
	VCSAG_LOGDBG_MSG(1, "LocalVV :: $sync_line[0] Sync Status is $self->{'sync_status'}.", 12, "$sync_line[0]", "$self->{'sync_status'}");
	return $ret;
}

#get_target_status : Get target Status
sub get_target_status {
	my $self = shift;
	
	my ($ret, $out) = $self->remote_copy_cmd($SHOWRCOPY_TARGET);
	if ($ret != 0) {
		return $ret;
        }

	my @out = split('\n', $out);
        while (@out) {
		$line = shift @out;
        	if ($line =~ /Target Information/) {
                	shift @out;
                	shift @out;
                	my $status_line = shift @out;
                	my (@status_line) =  split(" ", $status_line);
               		$self->{'target_status'} = $status_line[3];
               		last;
        	}
	}
	
	if ($self->{'target_status'} eq "") {
		VCSAG_LOG_MSG("E", "Failed to Initilize target status. Please verify Remote Copy configuration.", 13);
		$ret = 1;
	} else {
		VCSAG_LOGDBG_MSG(1, "Target $self->{'target'} is \"$self->{'target_status'}\".", 14, "$self->{'target'}", "$self->{'target_status'}" );
	}
	return $ret;
}


#check whether failover is required.
sub failover_required {
	my $self = shift;
	my $ret = 0;

	VCSAG_LOGDBG_MSG(1, "$self->{'role'}:: $self->{'group_status'} :: $self->{'sync_status'}.", 15, "$self->{'role'}", "$self->{'group_status'}", "$self->{'sync_status'}");
	
	#if ($self->{'role'} ne "Secondary") {
	#	VCSAG_LOGDBG_MSG(1, "Group role is $self->{'role'}. Failover is not required.", 16, "$self->{'role'}");
	#	$ret = 1;
	#}
	if (($self->{'group_status'} ne "Stopped") || ($self->{'sync_status'} ne "Stopped")) {
		VCSAG_LOGDBG_MSG(1, "Group status and Sync status should be in Stopped state. Failover is not possible.", 17);
                $ret = 1;
        }
	#Target status shall be Failed??
	if ($self->{'target_status'} ne "failed") {
                VCSAG_LOGDBG_MSG(1, "Target is not failed state. Failover is not possible.", 18);
                $ret = 1;
        }
	return $ret;
}

sub failover {
	my $self = shift;
	my ($ret, $out);
		
	if ($self->{'role'} eq "Secondary") {
		($ret, $out) = $self->remote_copy_cmd($FAILOVER);
	} elsif ($self->{'role'} eq "Secondary-Rev") {
		($ret, $out) = $self->remote_copy_cmd($FAILOVER_SEC_REV);
	}

	if ($ret == 0) {
		 VCSAG_LOG_MSG("I", "Group $self->{'group'} is successfully failover to $self->{'storage_server'}. Since replication is in stopped state and target is also in failed state so  note that group role on remote storage server $self->{'target'} could be Primary or Primary-Rev. Recover operation might need to be run to restart the remote copy replication.",19, "$self->{'group'}", "$self->{'storage_server'}", "$self->{'target'}");
		$ret = 0;
	} else {
		VCSAG_LOG_MSG("E", "Group $self->{'group'} is Failed to failover to $self->{'storage_server'}.", 20, "$self->{'group'}", "$self->{'storage_server'}");
		$ret = 1;
	}
	return $ret;
}


sub is_restore_allowed {
	my $self = shift;
        my $ret = 0;

	if (($self->{'role'} eq "Primary-Rev") && ($self->{'target_status'} eq "ready")) {
		$ret = 0;
	} else {
		VCSAG_LOG_MSG("E", "Group role is not Primary-Rev or target is not ready. Restore is not allowed.", 21);
		$ret = 1;
	}
	#Resorce shall be offline??
	if ($self->is_resource_online() == 0) {
		VCSAG_LOG_MSG("E", "RemoteCopy Resource $self->{'resname'} is ONLINE in local cluster. Restore is not allowed.", 22, "$self->{'resname'}");
		$ret = 1;
	}
	return $ret;
}

sub restore {
	my $self = shift;
        my $ret = 0;
	
	$ret = $self->init_values();
	if ($ret == 0) {
		return 1;
	}

	$ret = $self->is_restore_allowed();
	if ($ret == 1) {
		#Resorce shall be offline??
		return 1;
	}
	my ($ret, $out) = $self->remote_copy_cmd($RESTORE);

        if ($ret == 0) {
                VCSAG_LOG_MSG("I", "Group $self->{'group'} is successfully Restore to $self->{'storage_server'}.", 23, "$self->{'group'}", "$self->{'storage_server'}");
        } else {
                VCSAG_LOG_MSG("E", "Group $self->{'group'} is Failed to Restore to $self->{'storage_server'}.", 24, "$self->{'group'}", "$self->{'storage_server'}");
        }
        return $ret;
}


sub is_recover_allowed {
	my $self = shift;
        my $ret = 1;
	
	#How to check whether DR happened?
	if (($self->{'role'} eq "Primary-Rev") && ($self->{'group_status'} eq "Stopped") && ($self->{'target_status'} eq "ready")) {
                $ret = 0;
        }else {
		VCSAG_LOG_MSG("E", "Either Group role is not Primary-Rev or group status is not Stopped or target is not ready. Recover is not allowed.", 25);
		$ret = 1;
	}
	
	#Resorce shall be offline??
	if ($self->is_resource_online_remoteclus() == 0) {
		VCSAG_LOG_MSG("E", "RemoteCopy Resource $self->{'resname'} is ONLINE in remote cluster. Recover is not allowed.", 26, "$self->{'resname'}");
		$ret = 1;
	}
        return $ret;
}




sub recover {
        my $self = shift;
        my $ret = 0;

        $ret = $self->init_values();
        if ($ret == 0) {
                VCSAG_LOG_MSG("E", "Unable to Initilize. Exiting online.", 27);
                return 1;
        }

	$ret = $self->is_recover_allowed();
	if ($ret == 1) {
                #Resorce shall be offline??
                return 1;
        }
        my ($ret, $out) = $self->remote_copy_cmd($RECOVER);
	
	my ($ret1, $out1) = $self->remote_copy_cmd($STARTRCOPYGROUP);
	

        if (($ret == 0) && ($ret1 == 0)) {
                VCSAG_LOG_MSG("I", "Group $self->{'group'} is successfully Recover to $self->{'storage_server'}.", 28, "$self->{'group'}", "$self->{'storage_server'}");
        } else {
                VCSAG_LOG_MSG("E", "Group $self->{'group'} is Failed to Recover to $self->{'storage_server'}.", 29, "$self->{'group'}", "$self->{'storage_server'}");
                $ret = 1;
        }
        return $ret;
}

sub is_revert_allowed {
        my $self = shift;
        my $ret = 1;

        #How to check whether DR happened?
        if (($self->{'role'} eq "Primary-Rev") && ($self->{'group_status'} eq "Stopped") && ($self->{'target_status'} eq "ready")) {
                $ret = 0;
        }else {
                VCSAG_LOG_MSG("E", "Either Group role is not Primary-Rev or group status is not Stopped or target is not ready. Revert is not allowed.", 30);
                $ret = 1;
        }
	
	#Resorce shall be offline??
	if ($self->is_resource_online() == 0) {
		VCSAG_LOG_MSG("E", "RemoteCopy Resource $self->{'resname'} is ONLINE in local cluster. Revert is not allowed.", 31, "$self->{'resname'}");
		$ret = 1;
	}
        return $ret;
}


sub startrcopy_remote {
	my $self = shift;
        my $ret = 0;

	my $ResID = $self->{'resname'};
        $ret = $self->get_remote_info($ResID, $RemClus, $RemNode);
        if ($ret == 0) {
       		VCSAG_LOG_MSG("I", "Executing startrcopygroup action entry point on Remote node $RemNode", 32, "$RemNode");

		if($RemClus == "") {
                	my $out  = `$HARES -action $ResID startrcopygroup -sys $RemNode`;
                	my @out = split('\n', $out);
                	print "Output :: $out";
		} else {
                	my $out  = `$HARES -action $ResID startrcopygroup -sys $RemNode -clust $RemClus`;
                	my @out = split('\n', $out);
                	print "Output :: $out";
		}
		
	} else {
		VCSAG_LOG_MSG("E", "Failed to get RUNNING global Remote cluster node.", 47);
	}  
	return $ret;
}

sub revert {
        my $self = shift;
        my $ret = 0;

        $ret = $self->init_values();
        if ($ret == 0) {
                VCSAG_LOG_MSG("E", "Unable to Initilize. Exiting online.", 33);
                return 1;
        }
	
	$ret = $self->is_revert_allowed();
	if ($ret == 1) {
		return 1;
	}

        my ($ret, $out) = $self->remote_copy_cmd($REVERT);
        my @out = split('\n', $out);
	#Remote group shall be started.
	my $ret1 = $self->startrcopy_remote();

        if ($ret == 0) {
                VCSAG_LOG_MSG("I", "Group $self->{'group'} is successfully Recover to $self->{'storage_server'}.", 34, "$self->{'group'}", "$self->{'storage_server'}");
        } else {
                VCSAG_LOG_MSG("E", "Group $self->{'group'} is Failed to Recover to $self->{'storage_server'}.", 35, "$self->{'group'}", "$self->{'storage_server'}");
                $ret = 1;
        }
        return $ret;
}


sub migrate {
	my $self = shift;
	my $ret = 0;
	my $out = "";
	my $ret1 = 0;
	my $out1 = "";
	my ($ResID, $RemClus, $RemNode);
	
	#Need to remove this if and else condition
	if (($self->{'role'} eq "Secondary") || ($self->{'role'} eq "Secondary-Rev")) {
		
		#Its planed migration just reverse the role of replication.
		($ret, $out) = $self->remote_copy_cmd($MIGRATE);
		
		if ((!$ret) && ($self->{'swap_roles'} == 1)) {
			if ($self->{'role'} eq "Secondary") {
				VCSAG_LOG_MSG("I", "SwapRoles attribute is set, Swaping natural direction of the replication.", 36);
				($ret1, $out1) = $self->remote_copy_cmd($SWAPROLES);
			} else {
				VCSAG_LOG_MSG("N", "SwapRoles attribute is set but Group role is Secondary-Rev, Natural replication direction does not swapped.", 37);
			}
		}
        	if (($ret == 0 ) && ($ret1 == 0)) {
			my ($ret2, $out2) = $self->remote_copy_cmd($STARTRCOPYGROUP);
			if ($ret2 != 0) {
				 if (($out2 =~ /currently being promoted/) or
                                        ($out2 =~ /could be retried later/)) {
				
					$ret2 = $self->wait_to_promote();
				}	 
				if ($ret2 != 0 ){
				VCSAG_LOG_MSG("E", "Group $self->{'group'} Role changed to Ptimary/Primary-Rev but failed start remote copy group.Start remote copy group manually then run online again. Administative action is required", 38, "$self->{'group'}");  
				}
			}
		}

        	if (($ret == 0 ) && ($ret1 == 0) && ($ret2 == 0)){
                	VCSAG_LOG_MSG("I", "Group $self->{'group'} is successfully failover to $self->{'storage_server'}.", 39, "$self->{'group'}", "$self->{'storage_server'}");
        		$ret = 0;
		} else {
                	VCSAG_LOG_MSG("E", "Group $self->{'group'} is Failed to failover to $self->{'storage_server'}.", 40, "$self->{'group'}", "$self->{'storage_server'}");
                	$ret = 1;
        	}
		
	}elsif ($self->{'role'} eq "Secondary-Rev") {
                #Looks like Failed primary recoverd. Restore operation on remote side is required to restore orignal replication direction.
		my $ResID = $self->{'resname'};
		$ret = $self->get_remote_info($ResID, $RemClus, $RemNode);
		if ($ret == 0) {
			VCSAG_LOG_MSG("I", "Executing Restore action entry point on Remote node $RemNode", 22, "$RemNode");
			
			if($RemClus == "") { 
                        	my $out  = `$HARES -action $ResID restore -sys $RemNode`;
				my @out = split('\n', $out);
	                	print "Output :: $out";
			} else {
				my $out  = `$HARES -action $ResID restore -sys $RemNode -clust $RemClus`;
				my @out = split('\n', $out);
	                	print "Output :: $out";
			}

        	        $ret = $self->init_values();

                	if ($self->{'role'} eq "Primary") {
                        	VCSAG_LOG_MSG("I", "Group $self->{'group'} is successfully failover to $self->{'storage_server'}.", 39, "$self->{'group'}", "$self->{'storage_server'}");
                        	$ret = 0;
                	} else {
                        	VCSAG_LOG_MSG("E", "Group $self->{'group'} is Failed to failover to $self->{'storage_server'}.", 40, "$self->{'group'}", "$self->{'storage_server'}");
                        	$ret = 1;
                	}	
		} else {
			VCSAG_LOG_MSG("E", "Failed to get RUNNING global Remote cluster node.", 47);
		}
        }
	return $ret;
}

sub is_valid_primary_status {
	my $self = shift;
        my $ret = 0;
	
	if ($self->{'group_status'} eq "Started") {
		$ret = 0;
	} elsif ($self->{'group_status'} eq "Stopped") {
		VCSAG_LOG_MSG("W", "Group $self->{'group'} role is $self->{'role'} but Group status is : $self->{'group_status'}, Data is not getting replicated to Secondary side.", 42, "$self->{'group'}", "$self ->{'role'}", "$self->{'group_status'}");
		$ret = 0;
	} else {
		VCSAG_LOG_MSG("E", "Group $self->{'group'} role is $self->{'role'} but Invalid group status :: $self->{'group_status'}, Unable to bring resource online. Administrative intervention required.", 43, "$self->{'group'}", "$self ->{'role'}", "$self->{'group_status'}");
		$ret = 1;
	} 
	#elsif if ($self->{'sync_status'} eq "New") {
        #        VCSAG_LOG_MSG("E", "Invalid Sync status :: $self->{'sync_status'}", 9, "$self->{'sync_status'}");
        #        $ret = 1;
        #}
	 
	return $ret;
}
	

sub syncrcopy_remote {
	my $self = shift;
        my $ret = 0;

	my $ResID = $self->{'resname'};
        $ret = $self->get_remote_info($ResID, $RemClus, $RemNode);
        if ($ret == 0) {
       		VCSAG_LOG_MSG("I", "Executing startrcopygroup action entry point on Remote node $RemNode", 32, "$RemNode");

                if ($RemClus == "") {
			my $out  = `$HARES -action $ResID syncrcopy -sys $RemNode`;
			my @out = split('\n', $out);
	                print "Output :: $out";
		}else {
			my $out  = `$HARES -action $ResID syncrcopy -sys $RemNode -clust $RemClus`;
			my @out = split('\n', $out);
	                print "Output :: $out";
		}
                
		$ret = $?;
	        VCSAG_LOGDBG_MSG(1, "Remote action complete with return code : $ret Output : $out", 67, "$ret", "$out");
		if ($ret != 0) {
			#Remote action syncrcopy failed.
			VCSAG_LOG_MSG("E", "Failed to sync the group with output :: $out", 9, "$out");
			return $ret;
		}
		$ret = $self->get_sync_status();
        	if ($ret != 0) {
                	return $ret;
        	}
		if ($self->{'sync_status'} eq "Syncing") {
			$ret = $self->wait_for_group_sync();
		}
	} else {
		VCSAG_LOG_MSG("E", "Failed to get RUNNING global Remote cluster node.", 47);
	}
	return $ret;
}

sub is_valid_RemoteCopy_status {
	my $self = shift;
	my $ret = 0;

	if ($self->{'target_status'} eq "failing") {
		$ret = $self->wait_to_target_failed();
		if ($ret == 1) {
			VCSAG_LOG_MSG("E", "Target status is still $self->{'target_status'}. Please try after target either failed or ready.", 73, "$self->{'target_status'}");
			return 1;
		}
	}

	if ($self->is_valid_Sync_status() || $self->is_valid_group_status() || $self->is_valid_target_status()) {
		return 1;
	}
	if (($self->{'force_sync'} == 1 ) && ($self->{'mode'} eq "Periodic") && ($self->{'sync_status'} eq "Synced")) {
		#force sync primary to secondary
		$ret = $self->syncrcopy_remote();
	}
	return $ret;
} 

# Checks if Group is in valid sync status.
sub is_valid_Sync_status {
	my $self = shift;
	my $ret = 0;
	if (($self->{'sync_status'} eq "New") || ($self->{'sync_status'} eq "NotSynced")) {
		VCSAG_LOG_MSG("E", "Invalid Sync status :: $self->{'sync_status'}", 44, "$self->{'sync_status'}");
		$ret = 1;
	} elsif (($self->{'sync_status'} eq "Stale") && ($self->{'target_status'} ne "failed")) {
		VCSAG_LOG_MSG("E", "Group Sync status is stale and target is not failed.", 45);
		$ret = 1;
	} elsif ($self->{'sync_status'} eq "Syncing") { 
		if ($self->{'target_status'} eq "ready") {
			VCSAG_LOG_MSG("I", "Group $self->{'group'} is in Syncing state. Waiting Sync status to be Synced.", 46,"$self->{'group'}");
			$ret = $self->wait_for_group_sync();
		} else {
			VCSAG_LOG_MSG("E", "Group $self->{'group'} is in Syncing state but target status is not ready. Not waiting group to be Synced.", 47,"$self->{'group'}");
			$ret = 1;
		}
	}
	return $ret;
}

#check if group status is valid
sub is_valid_group_status {
	my $self = shift;
	my $ret = 0;
	if ($self->{'group_status'} eq "Started") {
		$ret = 0;
	} elsif ($self->{'group_status'} eq "Stopped") {
		if ($self->{'target_status'} eq "failed") {
			 VCSAG_LOGDBG_MSG(1,"Group is Stopped and Target status is $self->{'target_status'}.",48, "$self->{'target_status'}");
			$ret = 0;
		} else {
			VCSAG_LOG_MSG("E","Invalid Group status Stopped and Target status is $self->{'target_status'}.",49, "$self->{'target_status'}");
			$ret = 1;
		}
	} else {
		VCSAG_LOG_MSG("E", "Invalid group status  :: $self->{'group_status'} or Invalid Target status :: $self->{'target_status'}", 
			50, "$self->{'sync_status'}", "$self->{'target_status'}");
		$ret = 1;
	}
	return $ret;
}

sub is_valid_target_status {
	my $self = shift;
	my $ret = 0;
	if (($self->{'target_status'} eq "ready") || ($self->{'target_status'} eq "failed")) {
		$ret = 0;
	} else {
		 VCSAG_LOG_MSG("E", "Invalid Target status :: $self->{'target_status'}.", 51, "$self->{'target_status'}");
		$ret = 1;
	}
	return $ret;
}

##Executes the given command on the storage server
##RC >= 0, return code of the command
##RC < 0, cannot determine the return code of the command
sub remote_copy_cmd {
	my $self = shift;
        my $cmd = shift;
        my $ret;
        my $rc_cmd;
	
	if ($cmd eq $SHOWRCOPY) {
		$rc_cmd = "showrcopy groups $self->{'group'}";
	} elsif ($cmd eq $SHOWRCOPY_TARGET) {
		$rc_cmd = "showrcopy targets $self->{'target'}";
	} elsif ($cmd eq $FAILOVER) {
		$rc_cmd = "setrcopygroup failover -f -waittask $self->{'group'}";
	}elsif ($cmd eq $MIGRATE) {
		$rc_cmd = "setrcopygroup reverse -f -current -stopgroups -waittask $self->{'group'}"; #; startrcopygroup -wait $self->{'group'}";
	}elsif ($cmd eq $STARTRCOPYGROUP) {
		$rc_cmd = "startrcopygroup -wait $self->{'group'}";
	}elsif ($cmd eq $RESTORE) {
		$rc_cmd = "setrcopygroup restore -f -waittask $self->{'group'}";
	}elsif ($cmd eq $RECOVER) {
		$rc_cmd = "setrcopygroup recover -f -waittask $self->{'group'}";
	}elsif ($cmd eq $REVERT) {
		$rc_cmd = "setrcopygroup reverse -f -local -current -waittask $self->{'group'}";
	}elsif ($cmd eq $SWAPROLES) {
		$rc_cmd = "setrcopygroup reverse -f -natural -waittask -stopgroups $self->{'group'}";
	}elsif ($cmd eq $FAILOVER_SEC_REV) {
		$rc_cmd = "setrcopygroup reverse -f -local -current -waittask $self->{'group'}";
	}elsif ($cmd eq $SYNCRCOPY) {
		$rc_cmd = "syncrcopy -w $self->{'group'}";
	}elsif ($cmd eq $SHOWDATE) {
		$rc_cmd = "showdate";
	}elsif ($cmd =~ /^$SHOWHOST/) {
		$rc_cmd = $cmd;
	}
	
	my ($rc, $out) = $self->run_remote_copy_cmd($rc_cmd);
	if ($rc != 0) {
		VCSAG_LOG_MSG("E", "Remote Copy command :: [$rc_cmd] Failed with Output [$out].", 52, "$rc_cmd", "$out");
	}
	return ($rc, $out);
}

sub run_remote_copy_cmd {
	my $self = shift;
	my $cmd = shift;
	my $ret; 
        my $rpa_cmd;
	VCSAG_LOGDBG_MSG(1, "Executing Remote copy command :: [$cmd]", 53, "$cmd"); 
	
	if ($self->{'usecli'} == 0 ) {
	       $rc_cmd =  "\"" .$self->{'ssh_bin'}."\" -l " . $self->{'user'} . " -i \"" . $self->{'ssh_path_to_id_file'} . "\" "
				. $self->{'storage_server'} . " \"" . $cmd. " \"". " 2>&1";
	} else {
		$rc_cmd = "\"$self->{'cli'}\" $cmd -sys $self->{'storage_server'} -pwf  \"$self->{'passwordfile'}\" 2>&1";
	}
	my $out = `$rc_cmd`;
	
	$ret = $?;
	VCSAG_LOGDBG_MSG(1, "Command output [$out]", 54, "$out"); 
	VCSAG_LOGDBG_MSG(1, "Command exit code [$ret]", 55, "$ret"); 
	return ($ret, $out);
}


#check if node running
sub is_node_running {
	my $self = shift;
        my $RemNode = shift;
        my $ret = 1;
        open(RDSP1, "$ag_i18n_inc::HASYS -state |");
        while (<RDSP1>) {
                my @line = split();
                my ($Clus, $Node) = split(':', $line[0]);
                next if ($Node ne $RemNode);
                if ($line[2] eq "RUNNING") {
                        $ret = 0;
                }
                last;
        }
        # Make sure the pipe is empty
        while (<RDSP1>) {next;}
 	close RDSP1;
	return $ret;
}


# Function : get_remote_info
# Desc     : Retrieve the name of the remote node and cluster
# Input    : $ResID
# Output   : $RemClust
# Output   : $RemNode
# Returns  : Name of the local node

sub get_remote_info {
	my $self = shift;
        my ($ResID, $RemClus, $RemNode) = @_;
	$ResID = $self->{'resname'};
        open(RDSP, "$ag_i18n_inc::HARES -display $ResID |");
        if ($self->is_global_group($ResID)) {
                while (<RDSP>) {
                        next unless /^$ResID\s+State\s/;
                        next if /\slocalclus/;
                        my @line = split();
                        ($RemClus, $RemNode) = split(':', $line[2]);
                        $ret = $self->is_node_running($RemNode);
                        if ($ret == 1) {
                                VCSAG_LOG_MSG("I", "Remote node $RemNode is not RUNNING. Looping for next node.", 70, "$RemNode");
                                $RemClus = $RemNode = "";
                                next;
                        }
			last;
                }
       	# Make sure the pipe is empty
        	while (<RDSP>) {next;}
        	close RDSP;
        } else {
		#get local system name
		my (@fields) = split('\s+', `$ag_i18n_inc::HASYS -nodeid`);
		my $localsys = $fields[5];
		VCSAG_LOGDBG_MSG(1, "get_remote_info:local system name is : $localsys");

		#get group name
		my $cmd = "$ag_i18n_inc::HARES -value $self->{'resname'} Group";
		my ($grp) = split ('\n', `$cmd`);
		VCSAG_LOGDBG_MSG(1, "get_remote_info:VCS group name is : $grp");
			
		#get group's system list
		$cmd = "$ag_i18n_inc::HAGRP -value $grp SystemList";
		my ($syslist) = split ('\n', `$cmd`);
		VCSAG_LOGDBG_MSG(1, "get_remote_info:system list of group $grp : $syslist");
			
		my (@sysnames) = split('\s+', $syslist);

		foreach my $sysname (@sysnames) {
			#exclude priority order
			if($sysname =~ /^[0-9]/) {
				next;
			}

			#exclude local system
			if ($sysname =~ /$localsys/) {
				next;
			}

			VCSAG_LOGDBG_MSG(1, "get_remote_info:system from group's system list :  $sysname");

			#execute show host command to exlude the hosts connected to the local storage
			#my $cmd = $SHOWHOST . " " . $sysname;
			#my ($ret, $out) = $self->remote_copy_cmd($cmd);
			#VCSAG_LOGDBG_MSG(1, "get_remote_info: showhost command [$cmd] returned output $out");


			#if($ret == 0 && $out !~ /no hosts listed/) {
			#	next;
			#}

			#if the host is not connected to local storage server, find if it's role is Primary or Primary-Rev


		        $cmd = `$ag_i18n_inc::HASYS -state $sysname`;
			chomp($cmd);
			VCSAG_LOGDBG_MSG(1, "get_remote_info: system state of $sysname is $cmd");

			if ( $cmd eq "RUNNING" ){
				my $output  = `$ag_i18n_inc::HARES -action $ResID getremoteinfo -sys $sysname  2>&1`;
                        	VCSAG_LOGDBG_MSG(1, "get_remote_info: getremoteinfo action ep returned output : $output");
				my (@infolines) = split ('\n', $output);
				foreach my $infoline (@infolines) {
					if($infoline =~ /primary/i) {
						$RemNode = $sysname;
						VCSAG_LOGDBG_MSG(1, "get_remote_info: RDC remote node: [$RemNode]");
						last;
					}
				}
			}
			if ($RemNode ne "") {
				last;
			}
				
		}
        }
 	#print "$RemClus :: $RemNode \n";
        #return 1 if ( $RemClus eq "" && $RemNode eq "" );
	return 1 if ($RemNode eq "" );
        $_[1] = $RemClus;
        $_[2] = $RemNode;
        return 0;
}

#check if configured group is global group
sub is_global_group {
	my $self = shift;
        my ($ResID) =  @_;
        my $ret = 0;
         my $cmd = "$ag_i18n_inc::HARES -value $ResID Group";
	my $Grp = `$cmd`;
        chomp $Grp;
        $cmd = "$ag_i18n_inc::HAGRP -value $Grp  ClusterList";
	my $ClusterList = `$cmd`;
        chomp $ClusterList;

        if ( $ClusterList eq "" ) {
                VCSAG_LOG_MSG("I", "Remote Copy Service Group $Grp is not configured as global group.", 56, "$Grp");
                $ret = 0;
        } else {
                VCSAG_LOG_MSG("I", "RemoteCopy Service Group $Grp is configured as global group.", 57, "$Grp");
                $ret = 1;
        }
        return $ret;
}

#Check if resource is online in local cluster
sub is_resource_online {
	my $self = shift;
        my ($res) = $self->{'resname'};
        my $ret = 1;
	my (@st) = split('\n', `$ag_i18n_inc::HARES -state $res -clus localclus`);
	shift @st; #Remove first output line
	while (@st) {
	        my $line = shift @st;
        	my @state  = split(" ", $line);
        	if ($state[3] eq "ONLINE") {
                	$ret = 0;
			VCSAG_LOG_MSG("I", "RemoteCopy resource is ONLINE on System $state[2] in the local cluster.", 58, "$state[2]");
        	}
	}
	return $ret;

}

#Check if RemoteCopy group is ONLINE in Remote Cluster
sub is_resource_online_remoteclus {
	my $self = shift;
        my ($res) = $self->{'resname'};
        my $ret = 1;
	my (@st) = split('\n', `$ag_i18n_inc::HARES -state $res`);
	shift @st;
	while (@st) {
	        my $line = shift @st;
		next if ($line =~ /\slocalclus/);
        	my @state  = split(" ", $line);
        	if ($state[3] eq "ONLINE") {
                	$ret = 0;
			VCSAG_LOG_MSG("I", "RemoteCopy resource is ONLINE on System $state[2] on the Remote cluster.", 59, "$state[2]");
        	}
	}
	return $ret;
}

#
# This subroutine waits for a period of time for the state of the
# replication to move out of Syncing to Sync.
# The period of time that this routine waits is the remaining time
# left for online ep to timeout minus a grace period of one second.
# The subroutine sleeps 5 seconds, checks the status of replication.
# If it is still Syncing, then stays in
# the loop. Else get out of the loop and return from the function.
# Returns 0 on success and 1 on failure
#
sub wait_for_group_sync {
    my ($self) = @_;
    my ($oto) = split('\n', `$ag_i18n_inc::HATYPE -value RemoteCopy OnlineTimeout`);
    my $cur_time = time();
    my $remaining_time = ($oto - ($cur_time - $self->{'start_time'}));
    my $count = int($remaining_time / 6) - 3;
    VCSAG_LOGDBG_MSG(1, "Count :: $count Online timeout :: $oto", 60, $count, $oto);
    if ($count <= 0) {
        return 1;
    }

    my $i = 1;
    my $state = "";
    while ($i <  $count) {
        $ret = $self->get_sync_status();
        if ($ret != 0) {
               return 1;
        }
	if ($self->{'sync_status'} ne "Syncing"){
            	VCSAG_LOG_MSG("I","State of replication for $self->{'group'} changed to $self->{'sync_status'}.", 61, "$self->{'group'}", "$self->{'sync_status'}");
            	if ($self->{'sync_status'} eq "Synced") {
			VCSAG_LOG_MSG("I","Group $self->{'group'} Sync Status is Synced now. Proceeding to online the resource.", 62, "$self->{'group'}");
			return 0;
		}
        }
        # sleep 5 seconds
	VCSAG_LOG_MSG("I","Group $self->{'group'} Sync Status is Syncing. Waiting to Sync Status to be Synced..", 63, "$self->{'group'}"); 
        sleep (5);
        $i = $i + 1;
    }

    return 1;
}

sub wait_to_target_failed {
    my ($self) = @_;
    my ($oto) = split('\n', `$ag_i18n_inc::HATYPE -value RemoteCopy OnlineTimeout`);
    my $cur_time = time();
    my $remaining_time = ($oto - ($cur_time - $self->{'start_time'}));
    my $count = int($remaining_time / 6) - 3;
    VCSAG_LOGDBG_MSG(1, "Count :: $count Online timeout :: $oto", 60, $count, $oto);
    if ($count <= 0) {
        return 1;
    }

    my $i = 1;
    my $state = "";
    while ($i <  $count) {
        $ret = $self->get_target_status();
        if ($ret != 0) {
               return 1;
        }
        if ($self->{'target_status'} ne "failing") {
                VCSAG_LOG_MSG("I","target Status of $self->{'target'} changed to $self->{'target_status'}.", 71, "$self->{'target'}", "$self->{'target_status'}");
                return 0;
        }
        # sleep 5 seconds
        VCSAG_LOG_MSG("I", "$self->{'target'} target Status is failing. Waiting to target status to be failed..", 72, "$self->{'target'}");
        sleep (5);
        $i = $i + 1;
    }

    return 1;
}


sub wait_to_promote {
    my ($self) = @_;
    my ($oto) = split('\n', `$ag_i18n_inc::HATYPE -value RemoteCopy OnlineTimeout`);
    my $cur_time = time();
    my $remaining_time = ($oto - ($cur_time - $self->{'start_time'}));
    my $count = int($remaining_time / 6) - 1;
    VCSAG_LOGDBG_MSG(1, "Count :: $count Online timeout :: $oto", 60, $count, $oto);
    if ($count <= 0) {
        return 1;
    }
    my ($ret2, $out2);
    my $i = 1;
    my $state = "";
    while ($i <  $count) {
	($ret2, $out2) = $self->remote_copy_cmd($STARTRCOPYGROUP);
        if ($ret2 != 0) {
		if (($out2 =~ /currently being promoted/) or
                   ($out2 =~ /could be retried later/)) {
			VCSAG_LOG_MSG("I","Group $self->{'group'} is being promoted. Waiting for start the remote copy..", 66, "$self->{'group'}");
		} else {
			VCSAG_LOG_MSG("E", "Failed to promote volume. Group $self->{'group'} is not started", 64, "$self->{'group'}");
			return 1;
		}
	} else {
		VCSAG_LOG_MSG("I", "Volume is promoted and Group $self->{'group'} is Started succesfully.", 65, "$self->{'group'}");
		return 0;
	}
	
        # sleep 5 seconds
        sleep (5);
        $i = $i + 1;
    }
    return 1;
}

# Function to compute Replication specific RPO using InForm cli
sub Rep_Tech_RPO {
	my $ResName=shift;
	my ($ResName, @agent_args) = @_;
	VCSAG_SET_ENVS ($ResName, 20091);
        my $USERLANG=$ENV{"LANG"};
        $ENV{"LANG"}="C";

	my $res = RemoteCopyAgent->new($ResName, @agent_args);
	my $found = 0;
	my $sync_time = 0;
	my $last_sync_time = "";
	my $current_time = "";
	my ($ret, $out) = $res->remote_copy_cmd($SHOWRCOPY);
	if ($ret != 0) {
		return "Rep_Tech_RPO: Failed to get RemoteCopy group details. Cannot compute RPO.";
	}
	my @out = split('\n', $out);
	while (@out) {
                $line = shift @out;
		if ($line =~ /Group Information/) {
			shift @out;
                	my $status_line = shift @out;
                	my $role_line = shift @out;
			my (@role_line) =  split(" ", $role_line);
			my $group_status = $role_line[2];
			my $mode = $role_line[4];
			if (($mode eq "Sync") && ($group_status eq "Started")) {
				VCSAG_LOGDBG_MSG(1, "RemoteCopy group is configured in Sync mode and group is Started. returning 0 RPO.");	
				return "Rep_Tech_RPO: Replication RPO = 00:00:00 [hh:mm:ss] [LastSyncTime - CurrentTime]";
			}	
		}
		if ($line =~ /LocalVV/) {
                        $found = 1;
                        $line = shift @out;
                        if ($line eq "") {
                                VCSAG_LOG_MSG("E", "Virtual volumes are not configured for RemoteCopy group $self->{'group'}. Please Configure volumes for RemoteCopy group.", 10, "$self->{'group'}");
				return "Rep_Tech_RPO: Virtual volumes are not configured for RemoteCopy group. Cannot compute RPO.";
                        }
                }
                if (($found == 1) && ($line ne "")) {
                        my (@sync_line)  = split (" ", $line);
                        $last_sync_time = join (" ", $sync_line[5],  $sync_line[6],  $sync_line[7]);
			my $temp_sync_time = str2time($last_sync_time);
			if (($temp_sync_time < $sync_time) || ($sync_time == 0)) {
				VCSAG_LOGDBG_MSG(1, "Setting Sync time to $temp_sync_time.");
				$sync_time = $temp_sync_time;
			}
                }
        }
	#Get Current time of storage server.
	($ret, $out) = $res->remote_copy_cmd($SHOWDATE);
	my @out = split('\n', $out);
	$line = shift @out;
	$line = shift @out;
	if($line =~ m/\d+\s+(.*?)\(/) {
		$current_time = $1;
		chomp $current_time;
	}
	VCSAG_LOGDBG_MSG(1, "LastSyncTime = $sync_time and CurrentTime = $current_time");
	if (($sync_time == 0) or ($current_time eq "")) {
		VCSAG_LOG_MSG("E", "Failed to get LastSyncTime[$last_sync_time] or CurrentTime[$current_time]. Cannot compute RPO.", 75,"$last_sync_time","$current_time");
		return "Rep_Tech_RPO: Failed to get LastSyncTime or CurrentTime. Cannot compute RPO.";
	}
	my $curr_time = str2time($current_time);
	my $Rep_Tech_RPO = $curr_time - $sync_time;
	my $Replication_RPO=HADRSLA::time_format($Rep_Tech_RPO);
	return "Rep_Tech_RPO: Replication RPO = $Replication_RPO [hh:mm:ss] [LastSyncTime - CurrentTime]";
}


sub UpdateCurrentRPO {
	my $self = shift;
	my $computedrsla = shift;
	my $func = "UpdateCurrentRPO";
	my $ret = 1;
	my $rpo_out = "";
	if($computedrsla != 2) {
		return;
	}

	$self->serialize();
	my $cur_time = time();

	my ($current_rpo_val) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} CurrentRPO`);

	if ( $current_rpo_val eq "" ) {
		VCSAG_LOGDBG_MSG(1, "CurrentRPO attribute value is null, possible reason computing RPO first time.");

	} else {

		VCSAG_LOGDBG_MSG(1, "CurrentRPO attribute value   = $current_rpo_val", 55, $current_rpo_val);
		my($rpo_cal_time) = split(',' , $current_rpo_val);
		VCSAG_LOGDBG_MSG(1, "CurrentRPO attribute time value   = $rpo_cal_time", 54, $rpo_cal_time);


		VCSAG_LOGDBG_MSG(1, "Current time value   = $cur_time", 56, $cur_time);
		my $remain_time = $cur_time - $rpo_cal_time;
		VCSAG_LOGDBG_MSG(1, "Difference between current time and RPO calculate time = $remain_time", 57, $remain_time);
		if($remain_time <  240) {
#240 is equivalent to 4 min, 
#compute RPO logic should be executed with 5 min of interval.
#if value is set to 300 instead of 240 then there is chances of compute window exceeds 5min interval
			VCSAG_LOGDBG_MSG(1, "Time out is not yet reach, skip RPO calculation");
			return ;
		}
	}
	if ($self->{'usecli'} == 1) {
		VCSAG_LOGDBG_MSG(1, "Cli option is  not supported. only ssh is supported.");
#my $time = RemoteCopyAgent::get_remotecopy_rpo_cli($self->{'group'}, $self->{'storage_server'}, $self->{'passwordfile'},  $self->{'cli'},  $self->{'ssh_bin'} , $self->{'usecli'});
		return;	
	} else {
		($ret ,$rpo_out) = RPO::get_remotecopy_rpo_ssh($self->{'group'}, $self->{'storage_server'}, $self->{'ssh_bin'}, $self->{'ssh_path_to_id_file'}, $self->{'user'} );
	}

	if($ret) {
		VCSAG_LOGDBG_MSG(1, "RPO calculation failed. error[$rpo_out]", 58, $rpo_out);
		return ;
	}

	my $current_rpo = $cur_time . ",".$rpo_out;
	`$ag_i18n_inc::HARES -modify $self->{'resname'} CurrentRPO $current_rpo`;
	return ;
}

